import logging
import os
import shutil
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, List, Tuple

import pydantic
import requests
import yaml

try:
    from shared.util import BLOB_DOWNLOAD_TIMEOUT_SECS
except ModuleNotFoundError:
    from truss.templates.shared.util import BLOB_DOWNLOAD_TIMEOUT_SECS

try:
    import truss_transfer

    TRUSS_TRANSFER_AVAILABLE = True
except ImportError:
    TRUSS_TRANSFER_AVAILABLE = False

LAZY_DATA_RESOLVER_PATH = Path("/bptr/bptr-manifest")
NUM_WORKERS = 4
CACHE_DIR = Path("/cache/org/artifacts")
BASETEN_FS_ENABLED_ENV_VAR = "BASETEN_FS_ENABLED"

logger = logging.getLogger(__name__)


class Resolution(pydantic.BaseModel):
    url: str
    expiration_timestamp: int


class BasetenPointer(pydantic.BaseModel):
    """Specification for lazy data resolution for download of large files, similar to Git LFS pointers"""

    resolution: Resolution
    uid: str
    file_name: str
    hashtype: str
    hash: str
    size: int


class BasetenPointerManifest(pydantic.BaseModel):
    pointers: List[BasetenPointer]


class LazyDataResolver:
    """Deprecation warning: This class is deprecated and will be removed in a future release.

    Please use LazyDataResolverV2 instead (using the `truss_transfer` package).
    """

    def __init__(self, data_dir: Path):
        self._data_dir: Path = data_dir
        self._bptr_resolution: Dict[str, Tuple[str, str, int]] = _read_bptr_resolution()
        self._resolution_done = False
        self._uses_b10_cache = (
            os.environ.get(BASETEN_FS_ENABLED_ENV_VAR, "False") == "True"
        )

    def cached_download_from_url_using_requests(
        self, URL: str, hash: str, file_name: str, size: int
    ):
        """Download object from URL, attempt to write to cache and symlink to data directory if applicable, data directory otherwise.
        In case of failure, write to data directory
        """
        if self._uses_b10_cache:
            file_path = CACHE_DIR / hash
            if file_path.exists():
                try:
                    os.symlink(file_path, self._data_dir / file_name)
                    return
                except FileExistsError:
                    # symlink may already exist if the inference server was restarted
                    return

        # Streaming download to keep memory usage low
        resp = requests.get(
            URL, allow_redirects=True, stream=True, timeout=BLOB_DOWNLOAD_TIMEOUT_SECS
        )
        resp.raise_for_status()

        if self._uses_b10_cache:
            try:
                # Check whether the cache has sufficient space to store the file
                cache_free_space = shutil.disk_usage(CACHE_DIR).free
                if cache_free_space < size:
                    raise OSError(
                        f"Cache directory does not have sufficient space to save file {file_name}. Free space in cache: {cache_free_space}, file size: {size}"
                    )

                file_path.parent.mkdir(parents=True, exist_ok=True)
                with file_path.open("wb") as file:
                    shutil.copyfileobj(resp.raw, file)
                # symlink to data directory
                os.symlink(file_path, self._data_dir / file_name)
                return
            except FileExistsError:
                # symlink may already exist if the inference server was restarted
                return
            except OSError as e:
                logger.debug(
                    "Failed to save artifact to cache dir, saving to data dir instead. Error: %s",
                    e,
                )
                # Cache likely has no space left on device, break to download to data dir as fallback
                pass

        file_path = self._data_dir / file_name
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open("wb") as file:
            shutil.copyfileobj(resp.raw, file)

    def fetch(self):
        if self._resolution_done:
            return

        with ThreadPoolExecutor(NUM_WORKERS) as executor:
            futures = {}
            for file_name, (resolved_url, hash, size) in self._bptr_resolution.items():
                futures[
                    executor.submit(
                        self.cached_download_from_url_using_requests,
                        resolved_url,
                        hash,
                        file_name,
                        size,
                    )
                ] = file_name
            for future in as_completed(futures):
                if future.exception():
                    file_name = futures[future]
                    raise RuntimeError(f"Download failure for file {file_name}")
        self._resolution_done = True


class LazyDataResolverV2:
    def __init__(self, data_dir: Path):
        self._data_dir: Path = data_dir

    def fetch(self) -> str:
        if not LAZY_DATA_RESOLVER_PATH.is_file():
            logger.warning("Lazy data resolver path does not exist.")
            return ""
        return truss_transfer.lazy_data_resolve(str(self._data_dir))


def _read_bptr_resolution() -> Dict[str, Tuple[str, str, int]]:
    if not LAZY_DATA_RESOLVER_PATH.is_file():
        return {}
    bptr_manifest = BasetenPointerManifest(
        **yaml.safe_load(LAZY_DATA_RESOLVER_PATH.read_text())
    )
    resolution_map = {}
    for bptr in bptr_manifest.pointers:
        if bptr.resolution.expiration_timestamp < int(
            datetime.now(timezone.utc).timestamp()
        ):
            raise RuntimeError("Baseten pointer lazy data resolution has expired")
        resolution_map[bptr.file_name] = bptr.resolution.url, bptr.hash, bptr.size
    return resolution_map



from __future__ import annotations

import json
import logging
import os
import warnings
from enum import Enum
from typing import Annotated, Any, Dict, Literal, Optional

from huggingface_hub.errors import HFValidationError
from huggingface_hub.utils import validate_repo_id
from pydantic import (
    BaseModel,
    PydanticDeprecatedSince20,
    StringConstraints,
    model_validator,
    validator,
)

logger = logging.getLogger(__name__)
# Suppress Pydantic V1 warnings, because we have to use it for backwards compat.
warnings.filterwarnings("ignore", category=PydanticDeprecatedSince20)

ENGINE_BUILDER_TRUSS_RUNTIME_MIGRATION = (
    os.environ.get("ENGINE_BUILDER_TRUSS_RUNTIME_MIGRATION", "False") == "True"
)


class TrussTRTLLMModel(str, Enum):
    ENCODER = "encoder"
    DECODER = "decoder"
    # auto migrated settings
    PALMYRA = "palmyra"
    QWEN = "qwen"
    LLAMA = "llama"
    MISTRAL = "mistral"
    DEEPSEEK = "deepseek"
    # deprecated workflow
    WHISPER = "whisper"


class TrussTRTLLMQuantizationType(str, Enum):
    NO_QUANT = "no_quant"
    WEIGHTS_ONLY_INT8 = "weights_int8"
    WEIGHTS_KV_INT8 = "weights_kv_int8"
    WEIGHTS_ONLY_INT4 = "weights_int4"
    WEIGHTS_INT4_KV_INT8 = "weights_int4_kv_int8"
    SMOOTH_QUANT = "smooth_quant"
    FP8 = "fp8"
    FP8_KV = "fp8_kv"


class TrussTRTLLMPluginConfiguration(BaseModel):
    paged_kv_cache: bool = True
    use_paged_context_fmha: bool = True
    use_fp8_context_fmha: bool = False


class TrussTRTQuantizationConfiguration(BaseModel):
    """Configuration for quantization of TRT models

    Args:
        calib_size (int, optional): Size of calibration dataset. Defaults to 1024.
            recommended to increase for production runs (e.g. 1536), or decrease e.g. to 256 for quick testing.
        calib_dataset (str, optional): Hugginface dataset to use for calibration. Defaults to 'cnn_dailymail'.
            uses split='train' and  quantized based on 'text' column.
        calib_max_seq_length (int, optional): Maximum sequence length for calibration. Defaults to 2048.
    """

    calib_size: int = 1024
    calib_dataset: str = "cnn_dailymail"
    calib_max_seq_length: int = 2048

    def __init__(self, **data):
        super().__init__(**data)
        self.validate_cuda_friendly("calib_size")
        self.validate_cuda_friendly("calib_max_seq_length")

    def validate_cuda_friendly(self, key):
        value = getattr(self, key)
        if value < 64 or value > 16384:
            raise ValueError(f"{key} must be between 64 and 16384, but got {value}")
        elif value % 64 != 0:
            raise ValueError(f"{key} must be a multiple of 64, but got {value}")


class CheckpointSource(str, Enum):
    HF = "HF"
    GCS = "GCS"
    LOCAL = "LOCAL"
    # REMOTE_URL is useful when the checkpoint lives on remote storage accessible via HTTP (e.g a presigned URL)
    REMOTE_URL = "REMOTE_URL"


class CheckpointRepository(BaseModel):
    source: CheckpointSource
    repo: str
    revision: Optional[str] = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.source == CheckpointSource.HF:
            self._validate_hf_repo_id()

    def _validate_hf_repo_id(self):
        try:
            validate_repo_id(self.repo)
        except HFValidationError as e:
            raise ValueError(
                f"HuggingFace repository validation failed: {str(e)}"
            ) from e


class TrussTRTLLMBatchSchedulerPolicy(str, Enum):
    MAX_UTILIZATION = "max_utilization"
    GUARANTEED_NO_EVICT = "guaranteed_no_evict"


class TrussSpecDecMode(str, Enum):
    DRAFT_EXTERNAL = "DRAFT_TOKENS_EXTERNAL"
    LOOKAHEAD_DECODING = "LOOKAHEAD_DECODING"


class TrussTRTLLMRuntimeConfiguration(BaseModel):
    kv_cache_free_gpu_mem_fraction: float = 0.9
    kv_cache_host_memory_bytes: Optional[int] = None
    enable_chunked_context: bool = True
    batch_scheduler_policy: TrussTRTLLMBatchSchedulerPolicy = (
        TrussTRTLLMBatchSchedulerPolicy.GUARANTEED_NO_EVICT
    )
    request_default_max_tokens: Optional[int] = None
    total_token_limit: int = 500000
    webserver_default_route: Optional[
        Literal["/v1/embeddings", "/rerank", "/predict"]
    ] = None


class TrussTRTLLMBuildConfiguration(BaseModel):
    base_model: TrussTRTLLMModel = TrussTRTLLMModel.DECODER
    max_seq_len: Optional[int] = None
    max_batch_size: int = 256
    max_num_tokens: int = 8192
    max_beam_width: int = 1
    max_prompt_embedding_table_size: int = 0
    checkpoint_repository: CheckpointRepository
    gather_all_token_logits: bool = False
    strongly_typed: bool = False
    quantization_type: TrussTRTLLMQuantizationType = (
        TrussTRTLLMQuantizationType.NO_QUANT
    )
    quantization_config: TrussTRTQuantizationConfiguration = (
        TrussTRTQuantizationConfiguration()
    )
    tensor_parallel_count: int = 1
    pipeline_parallel_count: int = 1
    plugin_configuration: TrussTRTLLMPluginConfiguration = (
        TrussTRTLLMPluginConfiguration()
    )
    num_builder_gpus: Optional[int] = None
    speculator: Optional[TrussSpeculatorConfiguration] = None
    lora_adapters: Optional[
        Dict[
            Annotated[str, StringConstraints(pattern=r"^[a-z0-9]+$")],
            CheckpointRepository,
        ]
    ] = None

    class Config:
        extra = "forbid"

    def __init__(self, **data):
        super().__init__(**data)
        self._validate_kv_cache_flags()
        self._validate_speculator_config()

    def model_post_init(self, __context):
        self._bei_specfic_migration()

    @validator("max_beam_width")
    def check_max_beam_width(cls, v: int):
        if isinstance(v, int):
            if v != 1:
                raise ValueError(
                    "max_beam_width greater than 1 is not currently supported"
                )
        return v

    @property
    def uses_lookahead_decoding(self) -> bool:
        return (
            self.speculator is not None
            and self.speculator.speculative_decoding_mode
            == TrussSpecDecMode.LOOKAHEAD_DECODING
        )

    @property
    def uses_draft_external(self) -> bool:
        return (self.speculator is not None) and (
            self.speculator.speculative_decoding_mode == TrussSpecDecMode.DRAFT_EXTERNAL
        )

    @property
    def uses_lora(self) -> bool:
        return self.lora_adapters is not None and len(self.lora_adapters) > 0

    def _bei_specfic_migration(self):
        """performs embedding specfic optimizations (no kv-cache, high batch size)"""
        if self.base_model == TrussTRTLLMModel.ENCODER:
            # Encoder specific settings
            if self.max_seq_len:
                logger.info(
                    f"Your setting of `build.max_seq_len={self.max_seq_len}` is not used for embedding models, "
                    "and only respected for SequenceClassification models. "
                    "Automatically inferred from the model repo config.json -> `max_position_embeddings`"
                )
            # delayed import, as it is not available in all environments [Briton]
            from truss.base.constants import BEI_REQUIRED_MAX_NUM_TOKENS

            if self.max_num_tokens < BEI_REQUIRED_MAX_NUM_TOKENS:
                if self.max_num_tokens != 8192:
                    # only warn if it is not the default value
                    logger.warning(
                        f"build.max_num_tokens={self.max_num_tokens}, upgrading to {BEI_REQUIRED_MAX_NUM_TOKENS}"
                    )
                self = self.model_copy(
                    update={"max_num_tokens": BEI_REQUIRED_MAX_NUM_TOKENS}
                )
            # set page_kv_cache and use_paged_context_fmha to false for encoder
            self.plugin_configuration.paged_kv_cache = False
            self.plugin_configuration.use_paged_context_fmha = False

            if "_kv" in self.quantization_type.value:
                raise ValueError(
                    "encoder does not have a kv-cache, therefore a kv specfic datatype is not valid"
                    f"you selected build.quantization_type {self.quantization_type}"
                )

    def _validate_kv_cache_flags(self):
        if not self.plugin_configuration.paged_kv_cache and (
            self.plugin_configuration.use_paged_context_fmha
            or self.plugin_configuration.use_fp8_context_fmha
        ):
            raise ValueError(
                "Using paged context fmha or fp8 context fmha requires requires paged kv cache"
            )
        if (
            self.plugin_configuration.use_fp8_context_fmha
            and not self.plugin_configuration.use_paged_context_fmha
        ):
            raise ValueError("Using fp8 context fmha requires paged context fmha")
        if (
            self.plugin_configuration.use_fp8_context_fmha
            and not self.quantization_type == TrussTRTLLMQuantizationType.FP8_KV
        ):
            raise ValueError("Using fp8 context fmha requires fp8 kv cache dtype")
        return self

    def _validate_speculator_config(self):
        if self.speculator:
            if self.max_batch_size and self.max_batch_size > 64:
                logger.warning(
                    "We recommend speculative decoding for lower load on your servers, e.g. with batch-sizes below 32"
                    "To get better auto-tuned kernels, we recommend capping the max_batch_size to a more reasonable number e.g. `max_batch_size=64` or `max_batch_size=32`"
                    "If you have high batch-sizes, speculative decoding may not be beneficial for total throughput."
                    "If you want to use speculative decoding on high load, tune the concurrency settings for more aggressive autoscaling on Baseten."
                )

            if not all(
                [
                    self.plugin_configuration.use_paged_context_fmha,
                    self.plugin_configuration.paged_kv_cache,
                ]
            ):
                raise ValueError(
                    "KV cache block reuse must be enabled for speculative decoding target model."
                )

            if self.uses_draft_external and self.speculator.build:
                logger.warning(
                    "Draft external mode is a advanced feature. If you encounter issues, feel free to contact us. You may also try lookahead decoding mode instead."
                )
                if (
                    self.tensor_parallel_count
                    != self.speculator.build.tensor_parallel_count
                ):
                    raise ValueError(
                        "Speculative decoding requires the same tensor parallelism for target and draft models."
                    )

    @property
    def max_draft_len(self) -> Optional[int]:
        if self.speculator:
            return self.speculator.num_draft_tokens
        return None


class TrussSpeculatorConfiguration(BaseModel):
    speculative_decoding_mode: TrussSpecDecMode = TrussSpecDecMode.DRAFT_EXTERNAL
    num_draft_tokens: Optional[int] = None
    checkpoint_repository: Optional[CheckpointRepository] = None
    runtime: TrussTRTLLMRuntimeConfiguration = TrussTRTLLMRuntimeConfiguration()
    build: Optional[TrussTRTLLMBuildConfiguration] = None
    lookahead_windows_size: Optional[int] = None
    lookahead_ngram_size: Optional[int] = None
    lookahead_verification_set_size: Optional[int] = None

    def __init__(self, **data):
        super().__init__(**data)
        self._validate_checkpoint()
        self._validate_spec_dec_mode()

    def _assert_draft_tokens(self):
        if self.num_draft_tokens > 2048 or self.num_draft_tokens < 0:
            if self.speculative_decoding_mode == TrussSpecDecMode.LOOKAHEAD_DECODING:
                reason = (
                    f"This is automatically calculated value of lookahead_windows_size={self.lookahead_windows_size}, "
                    f" lookahead_ngram_size={self.lookahead_ngram_size}, lookahead_verification_set_size={self.lookahead_verification_set_size}. "
                    f"Please lower any of them."
                )
            else:
                reason = "You set this value under speculator.num_draft_tokens"
            raise ValueError(
                f"num_draft_tokens must be less than or equal to 2048. But you requested num_draft_tokens={self.num_draft_tokens}. {reason}"
            )

    @staticmethod
    def lade_max_draft_len(
        windows_size: int, ngram_size: int, verification_set_size: int
    ) -> int:
        """calculate the maximum number of tokens with baseten lookahead algorithm:  https://github.com/NVIDIA/TensorRT-LLM/tree/main/examples/lookahead#overview"""
        return (0 if (ngram_size == 1) else ngram_size - 2) + (
            windows_size - 1 + verification_set_size
        ) * (ngram_size - 1)

    def _validate_spec_dec_mode(self):
        if self.speculative_decoding_mode == TrussSpecDecMode.DRAFT_EXTERNAL:
            if not self.num_draft_tokens:
                raise ValueError(
                    "Draft external mode requires num_draft_tokens to be set."
                )
        elif self.speculative_decoding_mode == TrussSpecDecMode.LOOKAHEAD_DECODING:
            if not all(
                [
                    self.lookahead_windows_size,
                    self.lookahead_ngram_size,
                    self.lookahead_verification_set_size,
                ]
            ):
                raise ValueError(
                    f"Lookahead decoding mode requires lookahead_windows_size, lookahead_ngram_size, lookahead_verification_set_size to be set. you set: {self}"
                )
            lade_num_draft_tokens = self.lade_max_draft_len(
                self.lookahead_windows_size,
                self.lookahead_ngram_size,
                self.lookahead_verification_set_size,
            )
            if not ENGINE_BUILDER_TRUSS_RUNTIME_MIGRATION:
                if (
                    self.num_draft_tokens
                    and self.num_draft_tokens != lade_num_draft_tokens
                ):
                    raise ValueError(
                        f"num_draft_tokens is automatically calculated based on lookahead_windows_size, lookahead_ngram_size, lookahead_verification_set_size. "
                        f"Please remove num_draft_tokens or set it to exactly {lade_num_draft_tokens}. You set it to {self.num_draft_tokens}."
                    )
                self.num_draft_tokens = lade_num_draft_tokens
                if self.num_draft_tokens > 512:
                    logger.warning(
                        f"Lookahead decoding mode generates up to {self.num_draft_tokens} speculative tokens per step and may have performance implications. "
                        "We recommend a simpler config, e.g. lookahead_windows_size=7, lookahead_ngram_size=5, lookahead_verification_set_size=3."
                    )
            else:
                # server side on engine-builder
                if not self.num_draft_tokens:
                    raise ValueError(
                        "num_draft_tokens is required in lookahead decoding mode but not set"
                    )
                if (
                    self.num_draft_tokens < lade_num_draft_tokens
                ):  # check that it has at least the required tokens. That way, it could have even higher at request time.
                    raise ValueError(
                        "num_draft_tokens is less than the calculated value based on lookahead_windows_size, lookahead_ngram_size, lookahead_verification_set_size"
                    )

        self._assert_draft_tokens()

    def _validate_checkpoint(self):
        if self.speculative_decoding_mode == TrussSpecDecMode.DRAFT_EXTERNAL and not (
            bool(self.checkpoint_repository) ^ bool(self.build)
        ):
            raise ValueError(
                "Speculative decoding requires exactly one of checkpoint_repository or build to be configured."
            )

    @property
    def resolved_checkpoint_repository(self) -> CheckpointRepository:
        if self.build:
            return self.build.checkpoint_repository
        elif self.checkpoint_repository:
            return self.checkpoint_repository
        else:
            raise ValueError(
                "Speculative decoding requires exactly one of checkpoint_repository or build to be configured."
            )


class TRTLLMConfiguration(BaseModel):
    runtime: TrussTRTLLMRuntimeConfiguration = TrussTRTLLMRuntimeConfiguration()
    build: TrussTRTLLMBuildConfiguration

    def model_post_init(self, __context):
        self.add_bei_default_route()
        self.chunked_context_fix()

    @model_validator(mode="before")
    @classmethod
    def migrate_runtime_fields(cls, data: Any) -> Any:
        extra_runtime_fields = {}
        valid_build_fields = {}
        if isinstance(data.get("build"), dict):
            for key, value in data.get("build").items():
                if key in TrussTRTLLMBuildConfiguration.__annotations__:
                    valid_build_fields[key] = value
                else:
                    if key in TrussTRTLLMRuntimeConfiguration.__annotations__:
                        logger.warning(f"Found runtime.{key}: {value} in build config")
                        extra_runtime_fields[key] = value
            if extra_runtime_fields:
                logger.warning(
                    f"Found extra fields {list(extra_runtime_fields.keys())} in build configuration, unspecified runtime fields will be configured using these values."
                    " This configuration of deprecated fields is scheduled for removal, please upgrade to the latest truss version and update configs according to https://docs.baseten.co/performance/engine-builder-config."
                )
                if data.get("runtime"):
                    data.get("runtime").update(
                        {
                            k: v
                            for k, v in extra_runtime_fields.items()
                            if k not in data.get("runtime")
                        }
                    )
                else:
                    data.update(
                        {"runtime": {k: v for k, v in extra_runtime_fields.items()}}
                    )
            data.update({"build": valid_build_fields})
            return data
        return data

    def chunked_context_fix(self: "TRTLLMConfiguration") -> "TRTLLMConfiguration":
        """check if there is an error wrt. runtime.enable_chunked_context"""
        if (
            self.runtime.enable_chunked_context
            and (self.build.base_model != TrussTRTLLMModel.ENCODER)
            and not (
                self.build.plugin_configuration.use_paged_context_fmha
                and self.build.plugin_configuration.paged_kv_cache
            )
        ):
            logger.warning(
                "If trt_llm.runtime.enable_chunked_context is True, then trt_llm.build.plugin_configuration.use_paged_context_fmha and trt_llm.build.plugin_configuration.paged_kv_cache should be True. "
                "Setting trt_llm.build.plugin_configuration.use_paged_context_fmha and trt_llm.build.plugin_configuration.paged_kv_cache to True."
            )
            self.build = self.build.model_copy(
                update={
                    "plugin_configuration": self.build.plugin_configuration.model_copy(
                        update={"use_paged_context_fmha": True, "paged_kv_cache": True}
                    )
                }
            )

        return self

    def add_bei_default_route(self):
        if (
            self.runtime.webserver_default_route is None
            and self.build.base_model == TrussTRTLLMModel.ENCODER
            and not ENGINE_BUILDER_TRUSS_RUNTIME_MIGRATION
        ):
            # attemp to set the best possible default route client side.
            try:
                from transformers import AutoConfig

                hf_cfg = AutoConfig.from_pretrained(
                    self.build.checkpoint_repository.repo,
                    revision=self.build.checkpoint_repository.revision,
                )
                # simple heuristic to set the default route
                is_sequence_classification = (
                    "ForSequenceClassification" in hf_cfg.architectures[0]
                )
                route = "/predict" if is_sequence_classification else "/v1/embeddings"
                self.runtime = self.runtime.model_copy(
                    update={"webserver_default_route": route}
                )
                logger.info(
                    f"Setting default route to {route} for your encoder, as the model is a "
                    + (
                        "SequenceClassification Model."
                        if is_sequence_classification
                        else "Embeddings model."
                    )
                )
            except Exception:
                # access error, or any other issue
                pass

    @property
    def requires_build(self):
        return self.build is not None

    # TODO(Abu): Replace this with model_dump(json=True)
    # when pydantic v2 is used here
    def to_json_dict(self, verbose=True):
        return json.loads(self.json(exclude_unset=not verbose))



from unittest.mock import patch

import pytest

from truss.base.constants import TRTLLM_MIN_MEMORY_REQUEST_GI
from truss.trt_llm.config_checks import (
    is_missing_secrets_for_trt_llm_builder,
    memory_updated_for_trt_llm_builder,
)
from truss.truss_handle.truss_handle import TrussHandle


@patch("truss.trt_llm.config_checks._is_model_public")
@pytest.mark.parametrize(
    "has_secret, is_model_public, expected_result",
    [
        (False, False, True),
        (False, True, False),
        (True, False, False),
        (True, True, False),
    ],
)
def test_is_missing_secrets_for_trt_llm_builder(
    _is_model_public_mock,
    has_secret,
    is_model_public,
    expected_result,
    custom_model_trt_llm,
):
    _is_model_public_mock.return_value = is_model_public
    handle = TrussHandle(custom_model_trt_llm)
    if has_secret:
        handle.add_secret("hf_access_token")
    assert is_missing_secrets_for_trt_llm_builder(handle) == expected_result


def test_check_and_update_memory_for_trt_llm_builder(custom_model_trt_llm):
    handle = TrussHandle(custom_model_trt_llm)
    assert memory_updated_for_trt_llm_builder(handle)
    assert handle.spec.memory == f"{TRTLLM_MIN_MEMORY_REQUEST_GI}Gi"
    assert handle.spec.memory_in_bytes == TRTLLM_MIN_MEMORY_REQUEST_GI * 1024**3



from pathlib import Path
from typing import List, Optional

from truss.base.constants import CONFIG_FILE
from truss.base.truss_config import TrussConfig
from truss.truss_handle.patch.custom_types import TrussSignature
from truss.truss_handle.patch.dir_signature import directory_content_signature


def calc_truss_signature(
    truss_dir: Path, ignore_patterns: Optional[List[str]] = None
) -> TrussSignature:
    content_signature = directory_content_signature(truss_dir, ignore_patterns)
    config_path = truss_dir / CONFIG_FILE
    with (config_path).open("r") as config_file:
        config = config_file.read()
    requirements = TrussConfig.load_requirements_file_from_filepath(config_path)
    return TrussSignature(
        content_hashes_by_path=content_signature,
        config=config,
        requirements_file_requirements=requirements,
    )



import os
from unittest.mock import patch

import requests_mock

from truss.base.truss_config import ExternalData
from truss.util.download import download_external_data

TEST_DOWNLOAD_URL = "http://example.com/some-download-url"


def test_download(tmp_path):
    mocked_download_content = b"mocked content"
    with patch.dict(os.environ, {}), requests_mock.Mocker() as m:
        m.get(TEST_DOWNLOAD_URL, content=mocked_download_content)
        external_data = ExternalData.from_list(
            [{"local_data_path": "foo", "url": TEST_DOWNLOAD_URL}]
        )
        download_external_data(external_data=external_data, data_dir=tmp_path)

        with open(tmp_path / "foo", "rb") as f:
            content = f.read()

        assert content == mocked_download_content


def test_download_into_nested_subdir(tmp_path):
    mocked_download_content = b"mocked content"
    with patch.dict(os.environ, {}), requests_mock.Mocker() as m:
        m.get(TEST_DOWNLOAD_URL, content=mocked_download_content)
        external_data = ExternalData.from_list(
            [{"local_data_path": "foo/bar/baz", "url": TEST_DOWNLOAD_URL}]
        )
        download_external_data(external_data=external_data, data_dir=tmp_path)

        with open(tmp_path / "foo" / "bar" / "baz", "rb") as f:
            content = f.read()

        assert content == mocked_download_content



import base64

import requests
import sigint_patch

# We patch TensorRT-LLM v0.11 to fix an issue where they set a SIGINT handler
# which breaks because we do not use the main thread.
sigint_patch.patch()
from whisper_trt import WhisperModel
from whisper_trt.custom_types import WhisperResult


class Model:
    def __init__(self, **kwargs):
        self._data_dir = kwargs["data_dir"]
        self._secrets = kwargs["secrets"]
        self._model = None

    def load(self):
        self._model = WhisperModel(str(self._data_dir), max_queue_time=0.050)

    def preprocess(self, request: dict):
        audio_base64 = request.get("audio")
        audio_url = request.get("url")

        if audio_base64 and audio_url:
            return {
                "error": "Only a base64 audio file OR a URL can be passed to the API, not both of them."
            }
        if not audio_base64 and not audio_url:
            return {
                "error": "Please provide either an audio file in base64 string format or a URL to an audio file."
            }

        binary_data = None

        if audio_base64:
            binary_data = base64.b64decode(audio_base64.encode("utf-8"))
        elif audio_url:
            resp = requests.get(audio_url)
            binary_data = resp.content
        return binary_data, request

    async def predict(self, preprocessed_request) -> WhisperResult:
        binary_data, request = preprocessed_request
        waveform = self._model.preprocess_audio(binary_data)
        return await self._model.transcribe(
            waveform, language="english", timestamps=True
        )



import abc
import concurrent.futures
import inspect
import json
import logging
import pathlib
import textwrap
import traceback
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Type,
    cast,
)

import requests
import tenacity
import watchfiles

from truss.local import local_config_handler
from truss.remote import remote_factory
from truss.remote.baseten import core as b10_core
from truss.remote.baseten import custom_types as b10_types
from truss.remote.baseten import error as b10_errors
from truss.remote.baseten import remote as b10_remote
from truss.remote.baseten import service as b10_service
from truss.truss_handle import truss_handle
from truss.util import log_utils
from truss.util import path as truss_path
from truss_chains import framework, private_types, public_types, utils
from truss_chains.deployment import code_gen

if TYPE_CHECKING:
    from rich import console as rich_console
    from rich import progress


def _get_ordered_dependencies(
    chainlets: Iterable[Type[private_types.ABCChainlet]],
) -> Iterable[private_types.ChainletAPIDescriptor]:
    """Gather all Chainlets needed and returns a topologically ordered list."""
    needed_chainlets: set[private_types.ChainletAPIDescriptor] = set()

    def add_needed_chainlets(chainlet: private_types.ChainletAPIDescriptor):
        needed_chainlets.add(chainlet)
        for chainlet_descriptor in framework.get_dependencies(chainlet):
            needed_chainlets.add(chainlet_descriptor)
            add_needed_chainlets(chainlet_descriptor)

    for chainlet_cls in chainlets:
        add_needed_chainlets(framework.get_descriptor(chainlet_cls))
    # Get dependencies in topological order.
    return [
        descr
        for descr in framework.get_ordered_descriptors()
        if descr in needed_chainlets
    ]


def _get_chain_root(entrypoint: Type[private_types.ABCChainlet]) -> pathlib.Path:
    # TODO: revisit how chain root is inferred/specified, current might be brittle.
    chain_root = pathlib.Path(inspect.getfile(entrypoint)).absolute().parent
    logging.info(
        f"Using chain workspace dir: `{chain_root}` (files under this dir will "
        "be included as dependencies in the remote deployments and are importable)."
    )
    return chain_root


class ChainService(abc.ABC):
    """Handle for a deployed chain.

    A ``ChainService`` is created and returned when using ``push``. It
    bundles the individual services for each chainlet in the chain, and provides
    utilities to query their status, invoke the entrypoint etc.
    """

    _name: str
    _entrypoint_fake_json_data: Any

    def __init__(self, name: str):
        self._name = name
        self._entrypoint_fake_json_data = None

    @property
    def name(self) -> str:
        return self._name

    @property
    @abc.abstractmethod
    def status_page_url(self) -> str:
        """Link to status page on Baseten."""

    @property
    @abc.abstractmethod
    def run_remote_url(self) -> str:
        """URL to invoke the entrypoint."""

    @abc.abstractmethod
    def run_remote(self, json: Dict) -> Any:
        """Invokes the entrypoint with JSON data.

        Returns:
            The JSON response."""

    @abc.abstractmethod
    def get_info(self) -> list[b10_types.DeployedChainlet]:
        """Queries the statuses of all chainlets in the chain.

        Returns:
            List of ``DeployedChainlet``, ``(name, is_entrypoint, status, logs_url)``
            for each chainlet.
        """

    @property
    def entrypoint_fake_json_data(self) -> Any:
        """Fake JSON example data that matches the entrypoint's input schema.
        This property must be externally populated.

        Raises:
            ValueError: If fake data was not set.
        """
        if self._entrypoint_fake_json_data is None:
            raise ValueError("Fake data was not set.")
        return self._entrypoint_fake_json_data

    @entrypoint_fake_json_data.setter
    def entrypoint_fake_json_data(self, fake_data: Any) -> None:
        self._entrypoint_fake_json_data = fake_data


def _generate_chainlet_artifacts(
    options: private_types.PushOptions, entrypoint: Type[private_types.ABCChainlet]
) -> tuple[b10_types.ChainletArtifact, list[b10_types.ChainletArtifact], bool]:
    chain_root = _get_chain_root(entrypoint)
    entrypoint_artifact: Optional[b10_types.ChainletArtifact] = None
    dependency_artifacts: list[b10_types.ChainletArtifact] = []
    chainlet_display_names: set[str] = set()

    use_local_src = False
    if isinstance(options, private_types.PushOptionsLocalDocker):
        use_local_src = options.use_local_src

    has_engine_builder_chainlets = False

    for chainlet_descriptor in _get_ordered_dependencies([entrypoint]):
        if framework.is_engine_builder_chainlet(chainlet_descriptor.chainlet_cls):
            has_engine_builder_chainlets = True

        chainlet_display_name = chainlet_descriptor.display_name

        if chainlet_display_name in chainlet_display_names:
            raise public_types.ChainsUsageError(
                f"Chainlet names must be unique. Found multiple Chainlets with the name: '{chainlet_display_name}'."
            )

        chainlet_display_names.add(chainlet_display_name)

        # Since we are creating a distinct model for each deployment of the chain,
        # we add a random suffix.
        model_suffix = str(uuid.uuid4()).split("-")[0]
        model_name = f"{chainlet_display_name}-{model_suffix}"

        chainlet_dir = code_gen.gen_truss_chainlet(
            chain_root,
            options.chain_name,
            chainlet_descriptor,
            model_name,
            use_local_src,
        )
        artifact = b10_types.ChainletArtifact(
            truss_dir=chainlet_dir,
            name=chainlet_descriptor.name,
            display_name=chainlet_display_name,
        )

        is_entrypoint = chainlet_descriptor.chainlet_cls == entrypoint

        if is_entrypoint:
            assert entrypoint_artifact is None

            entrypoint_artifact = artifact
        else:
            dependency_artifacts.append(artifact)

    assert entrypoint_artifact is not None

    return entrypoint_artifact, dependency_artifacts, has_engine_builder_chainlets


@framework.raise_validation_errors_before
def push(
    entrypoint: Type[private_types.ABCChainlet],
    options: private_types.PushOptions,
    progress_bar: Optional[Type["progress.Progress"]] = None,
) -> Optional[ChainService]:
    entrypoint_artifact, dependency_artifacts, has_engine_builder_chainlets = (
        _generate_chainlet_artifacts(options, entrypoint)
    )
    if options.only_generate_trusses:
        return None
    if isinstance(options, private_types.PushOptionsBaseten):
        if has_engine_builder_chainlets and not options.publish:
            raise public_types.ChainsDeploymentError(
                "This chain contains engine builder chainlets. Development models are "
                "not supportd, push with `--publish`."
            )
        return _create_baseten_chain(
            options, entrypoint_artifact, dependency_artifacts, progress_bar
        )
    elif isinstance(options, private_types.PushOptionsLocalDocker):
        if has_engine_builder_chainlets:
            raise public_types.ChainsDeploymentError(
                "This chain contains engine builder chainlets. Running in local docker "
                "is not supported."
            )
        return _create_docker_chain(options, entrypoint_artifact, dependency_artifacts)
    else:
        raise NotImplementedError(options)


def push_debug_docker(
    entrypoint: Type[private_types.ABCChainlet], chain_name: str
) -> ChainService:
    options = private_types.PushOptionsLocalDocker(
        chain_name=chain_name, only_generate_trusses=False, use_local_src=True
    )
    return cast(ChainService, push(entrypoint, options))


# Docker ###############################################################################


class DockerChainletService(b10_service.TrussService):
    """This service is for Chainlets (not for Chains)."""

    def __init__(self, port: int, **kwargs):
        remote_url = f"http://localhost:{port}"

        super().__init__(remote_url, is_draft=False, **kwargs)

    def authenticate(self) -> Dict[str, str]:
        return {}

    def is_live(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    def is_ready(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    @property
    def logs_url(self) -> str:
        raise NotImplementedError()

    @property
    def predict_url(self) -> str:
        return f"{self._service_url}/v1/models/model:predict"

    def poll_deployment_status(self, sleep_secs: int = 1) -> Iterator[str]:
        raise NotImplementedError()


def _push_service_docker(
    truss_dir: pathlib.Path,
    chainlet_display_name: str,
    options: private_types.PushOptionsLocalDocker,
    port: int,
) -> None:
    th = truss_handle.TrussHandle(truss_dir)
    th.add_secret(public_types._BASETEN_API_SECRET_NAME, options.baseten_chain_api_key)
    th.docker_run(
        local_port=port,
        detach=True,
        wait_for_server_ready=True,
        network="host",
        container_name_prefix=chainlet_display_name,
        disable_json_logging=True,
    )


class DockerChainService(ChainService):
    _entrypoint_service: DockerChainletService

    def __init__(self, name: str, entrypoint_service: DockerChainletService) -> None:
        super().__init__(name)
        self._entrypoint_service = entrypoint_service

    @property
    def run_remote_url(self) -> str:
        """URL to invoke the entrypoint."""
        return self._entrypoint_service.predict_url

    def run_remote(self, json: Dict) -> Any:
        """Invokes the entrypoint with JSON data.

        Returns:
            The JSON response."""
        return self._entrypoint_service.predict(json)

    @property
    def status_page_url(self) -> str:
        """Not Implemented."""
        raise NotImplementedError()

    def get_info(self) -> list[b10_types.DeployedChainlet]:
        """Not Implemented."""
        raise NotImplementedError()


def _create_docker_chain(
    docker_options: private_types.PushOptionsLocalDocker,
    entrypoint_artifact: b10_types.ChainletArtifact,
    dependency_artifacts: list[b10_types.ChainletArtifact],
) -> DockerChainService:
    chainlet_artifacts = [*dependency_artifacts, entrypoint_artifact]
    chainlet_to_predict_url: Dict[str, Dict[str, str]] = {}
    chainlet_to_service: Dict[str, DockerChainletService] = {}
    for chainlet_artifact in chainlet_artifacts:
        port = utils.get_free_port()
        service = DockerChainletService(port)

        docker_internal_url = service.predict_url.replace(
            "localhost", "host.docker.internal"
        )
        chainlet_to_predict_url[chainlet_artifact.display_name] = {
            "predict_url": docker_internal_url
        }
        chainlet_to_service[chainlet_artifact.name] = service

        local_config_handler.LocalConfigHandler.set_dynamic_config(
            private_types.DYNAMIC_CHAINLET_CONFIG_KEY,
            json.dumps(chainlet_to_predict_url),
        )

        truss_dir = chainlet_artifact.truss_dir
        logging.info(
            f"Building Chainlet `{chainlet_artifact.display_name}` docker image."
        )
        _push_service_docker(
            truss_dir, chainlet_artifact.display_name, docker_options, port
        )
        logging.info(
            f"Pushed Chainlet `{chainlet_artifact.display_name}` as docker container."
        )
        logging.debug(
            "Internal model endpoint: "
            f"`{chainlet_to_predict_url[chainlet_artifact.display_name]}`"
        )

    return DockerChainService(
        docker_options.chain_name, chainlet_to_service[entrypoint_artifact.name]
    )


# Baseten ##############################################################################


class BasetenChainService(ChainService):
    _chain_deployment_handle: b10_core.ChainDeploymentHandleAtomic
    _remote: b10_remote.BasetenRemote

    def __init__(
        self,
        name: str,
        chain_deployment_handle: b10_core.ChainDeploymentHandleAtomic,
        remote: b10_remote.BasetenRemote,
    ) -> None:
        super().__init__(name)
        self._chain_deployment_handle = chain_deployment_handle
        self._remote = remote

    @property
    def run_remote_url(self) -> str:
        """URL to invoke the entrypoint."""

        handle = self._chain_deployment_handle

        return b10_service.URLConfig.invoke_url(
            hostname=handle.hostname,
            config=b10_service.URLConfig.CHAIN,
            entity_version_id=handle.chain_deployment_id,
            is_draft=handle.is_draft,
        )

    def run_remote(self, json_data: Dict) -> Any:
        """Invokes the entrypoint with JSON data.

        Returns:
            The JSON response."""
        headers = self._remote._auth_service.authenticate().header()
        response = requests.post(
            self.run_remote_url, json=json_data, headers=headers, stream=True
        )
        if response.status_code == 401:
            raise ValueError(
                f"Authentication failed with status code {response.status_code}"
            )

        if response.headers.get("transfer-encoding") == "chunked":
            # Case of streaming response, the backend does not set an encoding, so
            # manually decode to the contents to utf-8 here.
            def decode_content():
                for chunk in response.iter_content(
                    chunk_size=8192, decode_unicode=True
                ):
                    # Depending on the content-type of the response,
                    # iter_content will either emit a byte stream, or a stream
                    # of strings. Only decode in the bytes case.
                    if isinstance(chunk, bytes):
                        yield chunk.decode(
                            response.encoding or b10_service.DEFAULT_STREAM_ENCODING
                        )
                    else:
                        yield chunk

            return decode_content()

        parsed_response = response.json()

        if "error" in parsed_response:
            # In the case that the model is in a non-ready state, the response
            # will be a json with an `error` key.
            return parsed_response

        return response.json()

    @property
    def status_page_url(self) -> str:
        """Link to status page on Baseten."""
        return b10_service.URLConfig.status_page_url(
            self._remote.remote_url,
            b10_service.URLConfig.CHAIN,
            self._chain_deployment_handle.chain_id,
        )

    @tenacity.retry(
        stop=tenacity.stop_after_delay(300), wait=tenacity.wait_fixed(1), reraise=True
    )
    def get_info(self) -> list[b10_types.DeployedChainlet]:
        """Queries the statuses of all chainlets in the chain.

        Returns:
            List of ``DeployedChainlet`` for each chainlet."""
        return self._remote.get_chainlets(
            self._chain_deployment_handle.chain_deployment_id
        )


def _create_baseten_chain(
    baseten_options: private_types.PushOptionsBaseten,
    entrypoint_artifact: b10_types.ChainletArtifact,
    dependency_artifacts: list[b10_types.ChainletArtifact],
    progress_bar: Optional[Type["progress.Progress"]],
):
    logging.info(
        f"Pushing Chain '{baseten_options.chain_name}' to Baseten "
        f"(publish={baseten_options.publish}, environment={baseten_options.environment})."
    )
    remote_provider = cast(
        b10_remote.BasetenRemote,
        remote_factory.RemoteFactory.create(remote=baseten_options.remote),
    )

    if remote_provider.include_git_info or baseten_options.include_git_info:
        truss_user_env = b10_types.TrussUserEnv.collect_with_git_info(
            baseten_options.working_dir
        )
    else:
        truss_user_env = b10_types.TrussUserEnv.collect()

    _create_chains_secret_if_missing(remote_provider)

    chain_deployment_handle = remote_provider.push_chain_atomic(
        baseten_options.chain_name,
        entrypoint_artifact,
        dependency_artifacts,
        truss_user_env,
        publish=baseten_options.publish,
        environment=baseten_options.environment,
        progress_bar=progress_bar,
    )
    return BasetenChainService(
        baseten_options.chain_name, chain_deployment_handle, remote_provider
    )


def _create_chains_secret_if_missing(remote_provider: b10_remote.BasetenRemote) -> None:
    secrets_info = remote_provider.api.get_all_secrets()
    secret_names = {sec["name"] for sec in secrets_info["secrets"]}
    if public_types._BASETEN_API_SECRET_NAME not in secret_names:
        logging.info(
            "It seems you are using chains for the first time, since there "
            f"is no `{public_types._BASETEN_API_SECRET_NAME}` secret on baseten. "
            "Creating secret automatically."
        )
        remote_provider.api.upsert_secret(
            public_types._BASETEN_API_SECRET_NAME, remote_provider.api.auth_token.value
        )


# Watch / Live Patching ################################################################


def _create_watch_filter(root_dir: pathlib.Path):
    ignore_patterns = truss_path.load_trussignore_patterns_from_truss_dir(root_dir)

    def watch_filter(_: watchfiles.Change, path: str) -> bool:
        return not truss_path.is_ignored(pathlib.Path(path), ignore_patterns)

    logging.getLogger("watchfiles.main").disabled = True
    return ignore_patterns, watch_filter


def _handle_intercepted_logs(logs: list[str], console: "rich_console.Console"):
    if logs:
        formatted_logs = textwrap.indent("\n".join(logs), " " * 4)
        console.print(f"Intercepted logs from importing source code:\n{formatted_logs}")


def _handle_import_error(
    exception: Exception,
    console: "rich_console.Console",
    error_console: "rich_console.Console",
    stack_trace: Optional[str] = None,
):
    error_console.print(
        "Source files were changed, but pre-conditions for "
        "live patching are not given. Most likely there is a "
        "syntax error in the source files or names changed. "
        "Try to fix the issue and save the file. Error:\n"
        f"{textwrap.indent(str(exception), ' ' * 4)}"
    )
    if stack_trace:
        error_console.print(stack_trace)

    console.print(
        "The watcher will continue and if you can resolve the "
        "issue, subsequent patches might succeed.",
        style="blue",
    )


class _ModelWatcher:
    _source: pathlib.Path
    _model_name: str
    _remote_provider: b10_remote.BasetenRemote
    _ignore_patterns: list[str]
    _watch_filter: Callable[[watchfiles.Change, str], bool]
    _console: "rich_console.Console"
    _error_console: "rich_console.Console"

    def __init__(
        self,
        source: pathlib.Path,
        model_name: str,
        remote_provider: b10_remote.BasetenRemote,
        console: "rich_console.Console",
        error_console: "rich_console.Console",
    ) -> None:
        self._source = source
        self._model_name = model_name
        self._remote_provider = remote_provider
        self._console = console
        self._error_console = error_console
        self._ignore_patterns, self._watch_filter = _create_watch_filter(
            source.absolute().parent
        )

        dev_version = b10_core.get_dev_version(self._remote_provider.api, model_name)
        if not dev_version:
            raise b10_errors.RemoteError(
                "No development model found. Run `truss push` then try again."
            )

    def _patch(self) -> None:
        exception_raised = None
        with (
            log_utils.LogInterceptor() as log_interceptor,
            self._console.status(" Live Patching Model.\n", spinner="arrow3"),
        ):
            try:
                gen_truss_path = code_gen.gen_truss_model_from_source(self._source)
                return self._remote_provider.patch(
                    gen_truss_path,
                    self._ignore_patterns,
                    self._console,
                    self._error_console,
                )
            except Exception as e:
                exception_raised = e
            finally:
                logs = log_interceptor.get_logs()

        _handle_intercepted_logs(logs, self._console)
        if exception_raised:
            _handle_import_error(exception_raised, self._console, self._error_console)

    def watch(self) -> None:
        # Perform one initial patch at startup.
        self._patch()
        self._console.print("👀 Watching for new changes.", style="blue")

        # TODO(nikhil): Improve detection of directory structure, since right now
        # we assume a flat structure
        root_dir = self._source.absolute().parent
        for _ in watchfiles.watch(
            root_dir, watch_filter=self._watch_filter, raise_interrupt=False
        ):
            self._patch()
            self._console.print("👀 Watching for new changes.", style="blue")


class _Watcher:
    _source: pathlib.Path
    _entrypoint: Optional[str]
    _deployed_chain_name: str
    _remote_provider: b10_remote.BasetenRemote
    _chainlet_data: Mapping[str, b10_types.DeployedChainlet]
    _watch_filter: Callable[[watchfiles.Change, str], bool]
    _console: "rich_console.Console"
    _error_console: "rich_console.Console"
    _show_stack_trace: bool
    _included_chainlets: set[str]

    def __init__(
        self,
        source: pathlib.Path,
        entrypoint: Optional[str],
        name: Optional[str],
        remote: str,
        console: "rich_console.Console",
        error_console: "rich_console.Console",
        show_stack_trace: bool,
        included_chainlets: Optional[list[str]],
    ) -> None:
        self._source = source
        self._entrypoint = entrypoint
        self._console = console
        self._error_console = error_console
        self._show_stack_trace = show_stack_trace
        self._remote_provider = cast(
            b10_remote.BasetenRemote, remote_factory.RemoteFactory.create(remote=remote)
        )
        with framework.ChainletImporter.import_target(
            source, entrypoint
        ) as entrypoint_cls:
            self._deployed_chain_name = name or entrypoint_cls.__name__
            self._chain_root = _get_chain_root(entrypoint_cls)
            chainlet_names = set(
                desc.display_name
                for desc in _get_ordered_dependencies([entrypoint_cls])
            )

        if included_chainlets:
            if not_matched := (set(included_chainlets) - chainlet_names):
                raise public_types.ChainsDeploymentError(
                    "Requested to watch specific chainlets, but did not find "
                    f"{not_matched} among available chainlets {chainlet_names}."
                )
            self._included_chainlets = set(included_chainlets)
        else:
            self._included_chainlets = chainlet_names

        chain_id = b10_core.get_chain_id_by_name(
            self._remote_provider.api, self._deployed_chain_name
        )
        if not chain_id:
            raise public_types.ChainsDeploymentError(
                f"Chain `{chain_id}` was not found."
            )
        self._status_page_url = b10_service.URLConfig.status_page_url(
            self._remote_provider.remote_url, b10_service.URLConfig.CHAIN, chain_id
        )
        chain_deployment = b10_core.get_dev_chain_deployment(
            self._remote_provider.api, chain_id
        )
        if chain_deployment is None:
            raise public_types.ChainsDeploymentError(
                f"No development deployment was found for Chain `{chain_id}`. "
                "You cannot live-patch production deployments. Check the Chain's "
                f"status page for available deployments: {self._status_page_url}."
            )
        deployed_chainlets = self._remote_provider.get_chainlets(chain_deployment["id"])
        non_draft_chainlets = [
            chainlet.name for chainlet in deployed_chainlets if not chainlet.is_draft
        ]
        assert not (non_draft_chainlets), (
            "If the chain is draft, the oracles must be draft."
        )

        self._chainlet_data = {c.name: c for c in deployed_chainlets}
        self._assert_chainlet_names_same(chainlet_names)
        self._ignore_patterns, self._watch_filter = _create_watch_filter(
            self._chain_root
        )

    @property
    def _original_chainlet_names(self) -> set[str]:
        return set(self._chainlet_data.keys())

    def _assert_chainlet_names_same(self, new_names: set[str]) -> None:
        missing = self._original_chainlet_names - new_names
        added = new_names - self._original_chainlet_names
        if not (missing or added):
            return
        msg_parts = [
            "The deployed Chainlets and the Chainlets in the current workspace differ. "
            "Live patching is not possible if the set of Chainlet names differ."
        ]
        if missing:
            msg_parts.append(f"Chainlets missing in current workspace: {list(missing)}")
        if added:
            msg_parts.append(f"Chainlets added in current workspace: {list(added)}")

        raise public_types.ChainsDeploymentError("\n".join(msg_parts))

    def _code_gen_and_patch_thread(
        self, descr: private_types.ChainletAPIDescriptor
    ) -> tuple[b10_remote.PatchResult, list[str]]:
        with log_utils.LogInterceptor() as log_interceptor:
            # TODO: Maybe try-except code_gen errors explicitly.
            chainlet_dir = code_gen.gen_truss_chainlet(
                self._chain_root,
                self._deployed_chain_name,
                descr,
                self._chainlet_data[descr.display_name].oracle_name,
                use_local_src=False,
            )
            patch_result = self._remote_provider.patch_for_chainlet(
                chainlet_dir, self._ignore_patterns
            )
            logs = log_interceptor.get_logs()
        return patch_result, logs

    def _patch(self, executor: concurrent.futures.Executor) -> None:
        exception_raised = None
        stack_trace = ""
        with (
            log_utils.LogInterceptor() as log_interceptor,
            self._console.status(" Live Patching Chain.\n", spinner="arrow3"),
        ):
            # Handle import errors gracefully (e.g. if user saved file, but there
            # are syntax errors, undefined symbols etc.).
            try:
                with framework.ChainletImporter.import_target(
                    self._source, self._entrypoint
                ) as entrypoint_cls:
                    chainlet_descriptors = _get_ordered_dependencies([entrypoint_cls])
                    chain_root_new = _get_chain_root(entrypoint_cls)
                    assert chain_root_new == self._chain_root
                    self._assert_chainlet_names_same(
                        set(desc.display_name for desc in chainlet_descriptors)
                    )
                    future_to_display_name = {}
                    for chainlet_descr in chainlet_descriptors:
                        if chainlet_descr.display_name not in self._included_chainlets:
                            self._console.print(
                                f"⏩ Skipping patching `{chainlet_descr.display_name}`.",
                                style="grey50",
                            )
                            continue

                        future = executor.submit(
                            self._code_gen_and_patch_thread, chainlet_descr
                        )
                        future_to_display_name[future] = chainlet_descr.display_name
                    # Threads need to finish while inside the `import_target`-context.
                    done_futures = {
                        future_to_display_name[future]: future
                        for future in concurrent.futures.as_completed(
                            future_to_display_name
                        )
                    }
            except Exception as e:
                exception_raised = e
                stack_trace = traceback.format_exc()
            finally:
                logs = log_interceptor.get_logs()

        _handle_intercepted_logs(logs, self._console)
        if exception_raised:
            _handle_import_error(
                exception_raised,
                self._console,
                self._error_console,
                stack_trace=stack_trace if self._show_stack_trace else None,
            )
            return

        self._check_patch_results(done_futures)

    def _check_patch_results(
        self,
        display_name_to_done_future: Mapping[
            str, concurrent.futures.Future[tuple[b10_remote.PatchResult, list[str]]]
        ],
    ) -> None:
        has_errors = False
        for display_name, future in display_name_to_done_future.items():
            # It is not expected that code_gen_and_patch raises an exception, errors
            # should be handled by setting `b10_remote.PatchStatus`.
            # If an exception is raised anyway, it should bubble up the default way.
            patch_result, logs = future.result()
            if logs:
                formatted_logs = textwrap.indent("\n".join(logs), " " * 4)
                logs_output = f" [grey70]Intercepted logs:\n{formatted_logs}[grey70]"
            else:
                logs_output = ""

            if patch_result.status == b10_remote.PatchStatus.SUCCESS:
                self._console.print(
                    f"✅ Patched Chainlet `{display_name}`.{logs_output}", style="green"
                )
            elif patch_result.status == b10_remote.PatchStatus.SKIPPED:
                self._console.print(
                    f"💤 Nothing to do for Chainlet `{display_name}`.{logs_output}",
                    style="grey50",
                )
            else:
                has_errors = True
                self._error_console.print(
                    f"❌ Failed to patch Chainlet `{display_name}`. "
                    f"{patch_result.message}{logs_output}"
                )

        if has_errors:
            msg = (
                "Some Chainlets could not be live patched. See above error messages. "
                "The watcher will continue, and try patching new changes. However, the "
                "safest way to proceed and ensure a consistent state is to re-deploy "
                "the the entire development Chain."
            )
            self._error_console.print(msg)

    def watch(self) -> None:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Perform one initial patch at startup.
            self._patch(executor)
            self._console.print("👀 Watching for new changes.", style="blue")
            for _ in watchfiles.watch(
                self._chain_root, watch_filter=self._watch_filter, raise_interrupt=False
            ):
                self._patch(executor)
                self._console.print("👀 Watching for new changes.", style="blue")


@framework.raise_validation_errors_before
def watch(
    source: pathlib.Path,
    entrypoint: Optional[str],
    name: Optional[str],
    remote: str,
    console: "rich_console.Console",
    error_console: "rich_console.Console",
    show_stack_trace: bool,
    included_chainlets: Optional[list[str]],
) -> None:
    console.print(
        (
            "👀 Starting to watch for Chain source code and applying live patches "
            "when changes are detected."
        ),
        style="blue",
    )
    patcher = _Watcher(
        source,
        entrypoint,
        name,
        remote,
        console,
        error_console,
        show_stack_trace,
        included_chainlets,
    )
    patcher.watch()


def watch_model(
    source: pathlib.Path,
    model_name: str,
    remote_provider: b10_remote.TrussRemote,
    console: "rich_console.Console",
    error_console: "rich_console.Console",
):
    patcher = _ModelWatcher(
        source=source,
        model_name=model_name,
        remote_provider=cast(b10_remote.BasetenRemote, remote_provider),
        console=console,
        error_console=error_console,
    )
    patcher.watch()



from typing import Dict, List, Optional, Union

import pydantic

from truss.base import truss_config
from truss.shared import types


class SecretReference(types.SafeModel):
    name: str


class Compute(types.SafeModel):
    node_count: int = 1
    cpu_count: int = 1
    memory: str = "2Gi"
    accelerator: Optional[truss_config.AcceleratorSpec] = None

    def model_dump(self, *args, **kwargs):
        data = super().model_dump(*args, **kwargs)
        if self.accelerator and self.accelerator.accelerator:
            data["accelerator"] = {
                "accelerator": self.accelerator.accelerator.value,
                "count": self.accelerator.count,
            }
        return data


class Runtime(types.SafeModel):
    start_commands: List[str] = []
    environment_variables: Dict[str, Union[str, SecretReference]] = {}


class Image(types.SafeModel):
    base_image: str


class TrainingJob(types.SafeModel):
    image: Image
    compute: Compute = Compute()
    runtime: Runtime = Runtime()

    def model_dump(self, *args, **kwargs):
        data = super().model_dump(*args, **kwargs)
        data["compute"] = self.compute.model_dump()
        return data


class TrainingProject(types.SafeModel):
    name: str
    # TrainingProject is the wrapper around project config and job config. However, we exclude job
    # in serialization so just TrainingProject metadata is included in API requests.
    job: TrainingJob = pydantic.Field(exclude=True)



import os

from truss.util.env_vars import override_env_vars


def test_override_env_vars():
    os.environ["API_KEY"] = "original_key"

    with override_env_vars({"API_KEY": "new_key", "DEBUG": "true"}):
        assert os.environ["API_KEY"] == "new_key"
        assert os.environ["DEBUG"] == "true"

    assert os.environ["API_KEY"] == "original_key"
    assert "DEBUG" not in os.environ



import time
from typing import Any, List, Optional

from rich import console as rich_console

from truss.cli.logs.base_watcher import POLL_INTERVAL_SEC, LogWatcher
from truss.remote.baseten.api import BasetenApi

# NB(nikhil): When a job ends, we poll for this many seconds after to capture
# any trailing logs that contain information about errors.
JOB_TERMINATION_GRACE_PERIOD_SEC = 10

JOB_STARTING_STATES = ["TRAINING_JOB_CREATED", "TRAINING_JOB_DEPLOYING"]
JOB_RUNNING_STATES = ["TRAINING_JOB_RUNNING"]
JOB_ENDED_STATES = [
    "TRAINING_JOB_COMPLETED",
    "TRAINING_JOB_FAILED",
    "TRAINING_JOB_STOPPED",
]


class TrainingLogWatcher(LogWatcher):
    project_id: str
    job_id: str
    _poll_stop_time: Optional[int] = None
    _current_status: Optional[str] = None

    def __init__(
        self,
        api: BasetenApi,
        project_id: str,
        job_id: str,
        console: rich_console.Console,
    ):
        super().__init__(api, console)
        self.project_id = project_id
        self.job_id = job_id

    def _get_current_job_status(self) -> str:
        job = self.api.get_training_job(self.project_id, self.job_id)
        return job["training_job"]["current_status"]

    def before_polling(self) -> None:
        self._current_status = self._get_current_job_status()
        status_str = "Waiting for job to run, currently {current_status}..."
        with self.console.status(
            status_str.format(current_status=self._current_status), spinner="dots"
        ) as status:
            while self._current_status in JOB_STARTING_STATES:
                time.sleep(POLL_INTERVAL_SEC)
                self._current_status = self._get_current_job_status()
                status.update(status_str.format(current_status=self._current_status))

    def fetch_logs(
        self, start_epoch_millis: Optional[int], end_epoch_millis: Optional[int]
    ) -> List[Any]:
        return self.api.get_training_job_logs(
            self.project_id, self.job_id, start_epoch_millis, end_epoch_millis
        )

    def should_poll_again(self) -> bool:
        return self._current_status in JOB_RUNNING_STATES or self._poll_final_logs()

    def post_poll(self) -> None:
        self._current_status = self._get_current_job_status()
        self._maybe_update_poll_stop_time(self._current_status)

    def after_polling(self) -> None:
        if self._current_status == "TRAINING_JOB_COMPLETED":
            self.console.print("Training job completed successfully.", style="green")
        elif self._current_status == "TRAINING_JOB_FAILED":
            self.console.print("Training job failed.", style="red")
        elif self._current_status == "TRAINING_JOB_STOPPED":
            self.console.print("Training job stopped by user.", style="yellow")

    def _poll_final_logs(self):
        if self._poll_stop_time is None:
            return False

        return int(time.time()) <= self._poll_stop_time

    def _maybe_update_poll_stop_time(self, current_status: str) -> None:
        if current_status not in JOB_RUNNING_STATES and self._poll_stop_time is None:
            self._poll_stop_time = int(time.time()) + JOB_TERMINATION_GRACE_PERIOD_SEC



import contextlib
import importlib.util
import os
import pathlib
from typing import Iterator

from truss_train import definitions


@contextlib.contextmanager
def import_target(module_path: pathlib.Path) -> Iterator[definitions.TrainingProject]:
    module_name = module_path.stem
    if not os.path.isfile(module_path):
        raise ImportError(
            f"`{module_path}` is not a file. You must point to a python file where "
            f"the training configuration is defined."
        )

    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if not spec or not spec.loader:
        raise ImportError(f"Could not import `{module_path}`. Check path.")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    module_vars = (getattr(module, name) for name in dir(module))
    training_projects = [
        sym for sym in module_vars if isinstance(sym, definitions.TrainingProject)
    ]

    if len(training_projects) == 0:
        raise ValueError(f"No `{definitions.TrainingProject}` was found.")
    elif len(training_projects) > 1:
        raise ValueError(f"Multiple `{definitions.TrainingProject}`s were found.")

    yield training_projects[0]
